from dataclasses import dataclass, field
from pathlib import Path

from graphql import GraphQLSchema

from qenerate.core.preprocessor import GQLDefinition


@dataclass(eq=True, order=True)
class GeneratedFile:
    file: Path = field(compare=False)
    content: str

    def save(self) -> None:
        self.file.write_text(self.content)


@dataclass
class Fragment(GeneratedFile):
    definition: GQLDefinition
    import_path: str
    fragment_name: str
    class_name: str


class Plugin:
    def generate_operations(
        self,
        definitions: list[GQLDefinition],
        schema: GraphQLSchema,
        fragments: list[Fragment],
    ) -> list[GeneratedFile]:
        raise NotImplementedError

    def generate_fragments(
        self, definitions: list[GQLDefinition], schema: GraphQLSchema
    ) -> list[Fragment]:
        raise NotImplementedError



import re
from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum


class NamingCollisionStrategy(Enum):
    PARENT_CONTEXT = 1
    ENUMERATE = 2


@dataclass
class FeatureFlags:
    plugin: str
    gql_scalar_mappings: Mapping[str, str]
    collision_strategy: NamingCollisionStrategy = NamingCollisionStrategy.PARENT_CONTEXT


class FeatureFlagError(Exception):
    pass


class FeatureFlagParser:
    @staticmethod
    def plugin(definition: str) -> str:
        m = re.search(r"#\s*qenerate:\s*plugin\s*=\s*(\w+)\s*", definition)
        if not m:
            raise FeatureFlagError(
                "Missing valid qenerate plugin flag in query file: "
                "# qenerate: plugin=<plugin_id>"
            )
        return m.group(1)

    @staticmethod
    def naming_collision_strategy(definition: str) -> NamingCollisionStrategy:
        strategy = NamingCollisionStrategy.PARENT_CONTEXT
        m = re.search(
            r"#\s*qenerate:\s*naming_collision_strategy\s*=\s*(\w+)\s*", definition
        )
        if m:
            try:
                strategy = NamingCollisionStrategy[m.group(1)]
            except KeyError:
                raise FeatureFlagError(
                    f"Unknown naming_collision_strategy: {m.group(1)}"
                ) from None
        return strategy

    @staticmethod
    def custom_type_mapping(definition: str) -> dict[str, str]:
        mappings: dict[str, str] = {}
        m = re.findall(
            r"#\s*qenerate:\s*map_gql_scalar\s*=\s*(\w+)\s*->\s*(\w+)\s*",
            definition,
        )
        for groups in m:
            mappings[groups[0]] = groups[1]
        return mappings

    @staticmethod
    def parse(definition: str) -> FeatureFlags:
        return FeatureFlags(
            plugin=FeatureFlagParser.plugin(definition=definition),
            collision_strategy=FeatureFlagParser.naming_collision_strategy(
                definition=definition
            ),
            gql_scalar_mappings=FeatureFlagParser.custom_type_mapping(
                definition=definition
            ),
        )



# ruff: noqa: ANN401
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any

from graphql import (
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLSchema,
    OperationDefinitionNode,
    OperationType,
    Visitor,
    get_operation_ast,
    parse,
    validate,
    visit,
)

from qenerate.core.feature_flag_parser import FeatureFlagParser, FeatureFlags

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path


class AnonymousOperationError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(f"All operations must be named:\n{message}")


class GQLDefinitionType(Enum):
    QUERY = 1
    FRAGMENT = 2
    MUTATION = 3


@dataclass
class GQLDefinition:
    feature_flags: FeatureFlags
    source_file: Path
    kind: GQLDefinitionType
    definition: str
    name: str
    fragment_dependencies: set[str]


class DefinitionVisitor(Visitor):
    def __init__(self, source_file_path: Path, feature_flags: FeatureFlags) -> None:
        Visitor.__init__(self)
        self.definitions: list[GQLDefinition] = []
        self._feature_flags = feature_flags
        self._source_file_path = source_file_path
        self._stack: list[GQLDefinition] = []

    @staticmethod
    def _node_name(
        node: OperationDefinitionNode | FragmentDefinitionNode | FragmentSpreadNode,
    ) -> str:
        if not node.name:
            # TODO: proper error
            raise ValueError(f"{node} does not have a name")
        return node.name.value

    @staticmethod
    def _node_body(
        node: OperationDefinitionNode | FragmentDefinitionNode,
    ) -> str:
        if not node.loc:
            # TODO: proper error
            raise ValueError(f"{node} does not have loc set")
        start = node.loc.start_token.start
        end = node.loc.end_token.end
        return node.loc.source.body[start:end]

    def _add_definition(self) -> None:
        if self._stack:
            self.definitions.append(self._stack.pop())

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_: Any
    ) -> None:
        body = self._node_body(node)
        name = self._node_name(node)

        if node.operation == OperationType.QUERY:
            definition = GQLDefinition(
                kind=GQLDefinitionType.QUERY,
                definition=body,
                source_file=self._source_file_path,
                feature_flags=self._feature_flags,
                fragment_dependencies=set(),
                name=name,
            )
        elif node.operation == OperationType.MUTATION:
            definition = GQLDefinition(
                kind=GQLDefinitionType.MUTATION,
                definition=body,
                source_file=self._source_file_path,
                feature_flags=self._feature_flags,
                fragment_dependencies=set(),
                name=name,
            )
        else:
            # TODO: logger
            return
        self._stack.append(definition)

    def leave_operation_definition(self, *_: Any) -> None:
        self._add_definition()

    def enter_fragment_spread(self, node: FragmentSpreadNode, *_: Any) -> None:
        self._stack[-1].fragment_dependencies.add(self._node_name(node))

    def enter_fragment_definition(self, node: FragmentDefinitionNode, *_: Any) -> None:
        body = self._node_body(node)
        name = self._node_name(node)

        definition = GQLDefinition(
            kind=GQLDefinitionType.FRAGMENT,
            definition=body,
            source_file=self._source_file_path,
            feature_flags=self._feature_flags,
            fragment_dependencies=set(),
            name=name,
        )
        self._stack.append(definition)

    def leave_fragment_definition(self, *_: Any) -> None:
        self._add_definition()


class Preprocessor:
    @staticmethod
    def validate(definitions: Iterable[GQLDefinition], schema: GraphQLSchema) -> None:
        all_definitions = ""
        for definition in definitions:
            all_definitions += definition.definition + " "
            document_ast = parse(definition.definition)
            operation = get_operation_ast(document_ast)
            if operation and not operation.name:
                raise AnonymousOperationError(message=definition.definition)

        document_ast = parse(all_definitions)

        errors = validate(schema, document_ast)
        for error in errors:
            if "Fragment" in error.message and "is never used" in error.message:
                continue
            raise error

    @staticmethod
    def process_file(file_path: Path) -> list[GQLDefinition]:
        content = file_path.read_text(encoding="utf-8")
        feature_flags = FeatureFlagParser.parse(definition=content)
        document_ast = parse(content)
        visitor = DefinitionVisitor(
            feature_flags=feature_flags,
            source_file_path=file_path,
        )
        visit(document_ast, visitor)
        return visitor.definitions



# ruff: noqa: ANN401
from __future__ import annotations

import operator
from functools import reduce
from typing import TYPE_CHECKING, Any

from graphql import (
    FieldNode,
    FragmentDefinitionNode,
    FragmentSpreadNode,
    GraphQLOutputType,
    GraphQLScalarType,
    GraphQLSchema,
    InlineFragmentNode,
    OperationDefinitionNode,
    TypeInfo,
    TypeInfoVisitor,
    Visitor,
    parse,
    visit,
)

from qenerate.core.feature_flag_parser import FeatureFlags, NamingCollisionStrategy
from qenerate.core.plugin import (
    Fragment,
    GeneratedFile,
    Plugin,
)
from qenerate.core.preprocessor import GQLDefinition, GQLDefinitionType
from qenerate.core.unwrapper import Unwrapper, WrapperType
from qenerate.plugins.pydantic.mapper import (
    graphql_class_name_str_to_python,
    graphql_class_name_to_python,
    graphql_field_name_to_python,
    graphql_primitive_to_python,
)
from qenerate.plugins.pydantic.typed_ast import (
    BASE_CLASS_NAME,
    ParsedClassNode,
    ParsedFieldType,
    ParsedFragmentDefinitionNode,
    ParsedFragmentSpreadNode,
    ParsedInlineFragmentNode,
    ParsedNode,
    ParsedOperationNode,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

INDENT = "    "


def query_convenience_function(cls: str) -> str:
    return f"""
def query(query_func: Callable, **kwargs: Any) -> {cls}:
{INDENT}\"\"\"
{INDENT}This is a convenience function which queries and parses the data into
{INDENT}concrete types. It should be compatible with most GQL clients.
{INDENT}You do not have to use it to consume the generated data classes.
{INDENT}Alternatively, you can also mime and alternate the behavior
{INDENT}of this function in the caller.

{INDENT}Parameters:
{INDENT}{INDENT}query_func (Callable): Function which queries your GQL Server
{INDENT}{INDENT}kwargs: optional arguments that will be passed to the query function

{INDENT}Returns:
{INDENT}{INDENT}{cls}: queried data parsed into generated classes
{INDENT}\"\"\"
{INDENT}raw_data: dict[Any, Any] = query_func(DEFINITION, **kwargs)
{INDENT}return {cls}(**raw_data)
"""


def mutation_convenience_function(cls: str) -> str:
    return f"""
def mutate(mutation_func: Callable, **kwargs: Any) -> {cls}:
{INDENT}\"\"\"
{INDENT}This is a convenience function which executes a mutation and parses the response
{INDENT}into concrete types. It should be compatible with most GQL clients.
{INDENT}You do not have to use it to consume the generated data classes.
{INDENT}Alternatively, you can also mime and alternate the behavior
{INDENT}of this function in the caller.

{INDENT}Parameters:
{INDENT}{INDENT}mutation_func (Callable): Function which executes the mutation.
{INDENT}{INDENT}kwargs: Arguments that will be passed to the mutation function.
{INDENT}{INDENT}{INDENT}This must include the mutation parameters.

{INDENT}Returns:
{INDENT}{INDENT}{cls}: mutation response parsed into generated classes
{INDENT}\"\"\"
{INDENT}raw_data: dict[Any, Any] = mutation_func(DEFINITION, **kwargs)
{INDENT}return {cls}(**raw_data)
"""


class PydanticV1Error(Exception):
    pass


class FieldToTypeMatcherVisitor(Visitor):
    def __init__(
        self,
        schema: GraphQLSchema,
        type_info: TypeInfo,
        definition: GQLDefinition,
        feature_flags: FeatureFlags,
    ) -> None:
        Visitor.__init__(self)
        self.schema = schema
        self.type_info = type_info
        self.definition = definition
        self.feature_flags = feature_flags
        self.parsed = ParsedNode(
            parent=None,
            fields=[],
            parsed_type=ParsedFieldType(
                unwrapped_python_type="",
                wrapped_python_type="",
                is_primitive=False,
                enum_map={},
            ),
        )
        self.parent: ParsedNode | None = self.parsed
        self.deduplication_cache: set[str] = set()

    def enter_inline_fragment(self, node: InlineFragmentNode, *_: Any) -> None:
        graphql_type = self.type_info.get_type()
        if not graphql_type:
            raise ValueError(f"{node} does not have a graphql type")
        field_type = self._parse_type(graphql_type=graphql_type)
        current = ParsedInlineFragmentNode(
            fields=[],
            parent=self.parent,
            parsed_type=field_type,
        )
        if not self.parent:
            raise ValueError("Parent is None")
        self.parent.fields.append(current)
        self.parent = current

    def leave_inline_fragment(self, *_: Any) -> None:
        self.parent = self.parent.parent if self.parent else self.parent

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_: Any
    ) -> None:
        if not node.name:
            raise ValueError(f"{node} does not have a name")
        current = ParsedOperationNode(
            parent=self.parent,
            fields=[],
            operation_type=self.definition.kind,
            parsed_type=ParsedFieldType(
                unwrapped_python_type=node.name.value,
                wrapped_python_type=node.name.value,
                is_primitive=False,
                enum_map={},
            ),
        )
        if not self.parent:
            raise ValueError("Parent is None")
        self.parent.fields.append(current)
        self.parent = current

    def leave_operation_definition(self, *_: Any) -> None:
        self.parent = self.parent.parent if self.parent else self.parent

    def enter_fragment_definition(self, node: FragmentDefinitionNode, *_: Any) -> None:
        graphql_type = self.type_info.get_type()
        if not graphql_type:
            raise ValueError(f"{node} does not have a graphql type")
        field_type = self._parse_type(graphql_type=graphql_type)
        name = graphql_class_name_str_to_python(node.name.value)
        current = ParsedFragmentDefinitionNode(
            fields=[],
            parent=self.parent,
            parsed_type=field_type,
            class_name=name,
            fragment_name=node.name.value,
        )
        if not self.parent:
            raise ValueError("Parent is None")
        self.parent.fields.append(current)
        self.parent = current

    def leave_fragment_definition(self, *_: Any) -> None:
        self.parent = self.parent.parent if self.parent else self.parent

    def enter_fragment_spread(self, node: FragmentSpreadNode, *_: Any) -> None:
        fragment_name = graphql_class_name_str_to_python(node.name.value)
        field_type = ParsedFieldType(
            is_primitive=False,
            unwrapped_python_type=fragment_name,
            wrapped_python_type=fragment_name,
            enum_map={},
        )
        current = ParsedFragmentSpreadNode(
            fields=[],
            parent=self.parent,
            parsed_type=field_type,
        )
        if not self.parent:
            raise ValueError("Parent is None")
        self.parent.fields.append(current)
        self.parent = current

    def leave_fragment_spread(self, *_: Any) -> None:
        self.parent = self.parent.parent if self.parent else self.parent

    def enter_field(self, node: FieldNode, *_: Any) -> None:
        graphql_type = self.type_info.get_type()
        if not graphql_type:
            raise ValueError(f"{node} does not have a graphql type")
        field_type = self._parse_type(graphql_type=graphql_type)
        gql_key = node.alias.value if node.alias else node.name.value
        py_key = graphql_field_name_to_python(gql_key)
        current = ParsedClassNode(
            fields=[],
            parent=self.parent,
            parsed_type=field_type,
            py_key=py_key,
            gql_key=gql_key,
        )
        if not self.parent:
            raise ValueError("Parent is None")
        self.parent.fields.append(current)
        self.parent = current

    def leave_field(self, *_: Any) -> None:
        self.parent = self.parent.parent if self.parent else self.parent

    # Custom Functions
    def _parse_type(self, graphql_type: GraphQLOutputType) -> ParsedFieldType:
        unwrapper_result = Unwrapper.unwrap(graphql_type)
        unwrapped_type = self._to_python_type(unwrapper_result.inner_gql_type)
        is_primitive = unwrapper_result.is_primitive
        enum_map = unwrapper_result.enum_map
        wrapped_type = unwrapped_type
        for wrapper in reversed(unwrapper_result.wrapper_stack):
            if wrapper == WrapperType.LIST:
                wrapped_type = f"list[{wrapped_type}]"
            elif wrapper == WrapperType.OPTIONAL:
                wrapped_type = f"Optional[{wrapped_type}]"

        return ParsedFieldType(
            unwrapped_python_type=unwrapped_type,
            wrapped_python_type=wrapped_type,
            is_primitive=is_primitive,
            enum_map=enum_map,
        )

    def _to_python_type(self, graphql_type: GraphQLOutputType) -> str:
        if isinstance(graphql_type, GraphQLScalarType):
            return graphql_primitive_to_python(
                graphql_type=graphql_type,
                custom_mappings=self.feature_flags.gql_scalar_mappings,
            )
        cur = self.parent
        class_name = graphql_class_name_to_python(graphql_type=graphql_type)
        collision_enum_suffix = 2

        # handle name collisions
        while cur and cur.parent and class_name in self.deduplication_cache:
            # Enumerate (1,2,3, ...) if you see a collision
            if (
                self.feature_flags.collision_strategy
                == NamingCollisionStrategy.ENUMERATE
            ):
                if collision_enum_suffix == 2:  # noqa: PLR2004
                    class_name = f"{class_name}__{collision_enum_suffix}"
                else:
                    idx = class_name.rfind("_") - 1
                    class_name = f"{class_name[:idx]}__{collision_enum_suffix}"
                collision_enum_suffix += 1
            # By default, prefix with parent name if you see a collision
            else:
                class_name = f"{cur.parsed_type.unwrapped_python_type}_{class_name}"
            cur = cur.parent

        self.deduplication_cache.add(class_name)
        return class_name


class QueryParser:
    @staticmethod
    def parse(
        definition: GQLDefinition, schema: GraphQLSchema, feature_flags: FeatureFlags
    ) -> ParsedNode:
        document_ast = parse(definition.definition)
        type_info = TypeInfo(schema)
        visitor = FieldToTypeMatcherVisitor(
            schema=schema,
            type_info=type_info,
            definition=definition,
            feature_flags=feature_flags,
        )
        visit(document_ast, TypeInfoVisitor(type_info, visitor))
        return visitor.parsed


class PydanticBase(Plugin):
    HEADER: str
    IMPORTS: str
    CONF: str

    def _traverse(self, node: ParsedNode) -> str:
        """Traverse the AST.

        Pydantic doesnt play well with from __future__ import annotations
        --> order of class declaration is important:
        - post-order for non-inline fragment nodes, i.e., non-interface nodes
        - pre-order for nodes that implement an interface
        """
        result = ""
        for child in node.fields:
            if not isinstance(child, ParsedInlineFragmentNode):
                result = f"{result}{self._traverse(child)}"

        result = f"{result}{node.class_code_string()}"

        for child in node.fields:
            if isinstance(child, ParsedInlineFragmentNode):
                result = f"{result}{self._traverse(child)}"
        return result

    def generate_fragments(
        self, definitions: list[GQLDefinition], schema: GraphQLSchema
    ) -> list[Fragment]:
        """Render all fragments.

        Handle nested fragments, i.e., fragments which
        depend on other fragments. The current dependency resolving approach
        is brute-force, but will get the job done for most likely use-cases.

        @TODO: fragment dependency graph should be calculated in pre-processor.
        This function should ideally then be able to process fragments in order.
        """
        processed: dict[str, Fragment] = {}
        current_to_process: list[GQLDefinition] = definitions
        next_to_process: list[GQLDefinition] = []
        while len(processed) < len(definitions):
            for definition in current_to_process:
                # We make sure that all dependencies are already rendered
                # in order to obtain proper import lines
                has_unrendered_dependencies = reduce(
                    operator.or_,
                    [dep not in processed for dep in definition.fragment_dependencies],
                    False,  # noqa: FBT003
                )
                if has_unrendered_dependencies:
                    # not all dependencies rendered yet. It will
                    # probably happen later in the loop, so we skip
                    # this and will try again in the next iteration.
                    next_to_process.append(definition)
                    continue
                fragment_imports = self._fragment_imports(
                    definition=definition,
                    fragment_map=processed,
                )
                result = self.HEADER + self.IMPORTS
                if fragment_imports:
                    result += "\n"
                    result += fragment_imports
                result += f"\n\n\n{self.CONF}"
                qf = definition.source_file
                parser = QueryParser()
                ast = parser.parse(
                    definition=definition,
                    schema=schema,
                    feature_flags=definition.feature_flags,
                )
                fragment = ast.fields[0]
                if not isinstance(fragment, ParsedFragmentDefinitionNode):
                    continue
                result += self._traverse(ast)
                result += "\n"
                import_path = str(definition.source_file.with_suffix("")).replace(
                    "/", "."
                )
                rendered_fragment = Fragment(
                    definition=definition,
                    file=qf.with_suffix(".py"),
                    content=result,
                    class_name=fragment.class_name,
                    import_path=import_path,
                    fragment_name=fragment.fragment_name,
                )
                processed[rendered_fragment.fragment_name] = rendered_fragment

            if len(current_to_process) == len(next_to_process):
                # We found a cyclic dependency, i.e., not a single
                # fragment got rendered in this iteration
                raise PydanticV1Error("Cyclic fragment dependency detected")

            # lets try again to render all fragments that we skipped in this iteration
            current_to_process = next_to_process
            next_to_process = []

        return list(processed.values())

    @staticmethod
    def _fragment_imports(
        definition: GQLDefinition, fragment_map: Mapping[str, Fragment]
    ) -> str:
        imports = ""
        for dep in sorted(definition.fragment_dependencies):
            fragment = fragment_map[dep]
            imports += f"\nfrom {fragment.import_path} import {fragment.class_name}"
        return imports

    def _assemble_definition(
        self,
        definition: GQLDefinition,
        fragment_definitions: Mapping[str, GQLDefinition],
    ) -> set[str]:
        ans: set[str] = set()
        ans.add(definition.definition)
        for dep in sorted(definition.fragment_dependencies):
            ans = ans.union(
                self._assemble_definition(
                    definition=fragment_definitions[dep],
                    fragment_definitions=fragment_definitions,
                )
            )
        return ans

    def generate_operations(
        self,
        definitions: list[GQLDefinition],
        schema: GraphQLSchema,
        fragments: list[Fragment],
    ) -> list[GeneratedFile]:
        generated_files: list[GeneratedFile] = []
        fragment_map = {f.fragment_name: f for f in fragments}
        fragment_definitions = {f.fragment_name: f.definition for f in fragments}
        for definition in definitions:
            fragment_imports = self._fragment_imports(
                definition=definition,
                fragment_map=fragment_map,
            )

            result = self.HEADER + self.IMPORTS
            if fragment_imports:
                result += "\n"
                result += fragment_imports
            result += "\n\n\n"
            qf = definition.source_file
            assembled_definition = "\n\n".join(
                sorted(
                    self._assemble_definition(
                        definition=definition,
                        fragment_definitions=fragment_definitions,
                    )
                )
            )
            result += f'DEFINITION = """\n{assembled_definition}\n"""'
            result += f"\n\n\n{self.CONF}"
            parser = QueryParser()
            ast = parser.parse(
                definition=definition,
                schema=schema,
                feature_flags=definition.feature_flags,
            )
            result += self._traverse(ast)
            result += "\n\n"
            cls = ast.fields[0].parsed_type.unwrapped_python_type
            if definition.kind == GQLDefinitionType.QUERY:
                result += query_convenience_function(cls=f"{cls}QueryData")
            else:
                result += mutation_convenience_function(cls=f"{cls}MutationResponse")
            generated_files.append(
                GeneratedFile(file=qf.with_suffix(".py"), content=result)
            )

        return generated_files


class PydanticV1Plugin(PydanticBase):
    """Pydantic v1 plugin for Qenerate."""

    HEADER = (
        '"""\nGenerated by qenerate plugin=pydantic_v1. DO NOT MODIFY MANUALLY!\n"""\n'
    )

    IMPORTS = (
        "from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611\n"
        "from datetime import datetime  # noqa: F401 # pylint: disable=W0611\n"
        "from enum import Enum  # noqa: F401 # pylint: disable=W0611\n"
        "from typing import (  # noqa: F401 # pylint: disable=W0611\n"
        f"{INDENT}Any,\n"
        f"{INDENT}Optional,\n"
        f"{INDENT}Union,\n"
        ")\n"
        "\n"
        "from pydantic import (  # noqa: F401 # pylint: disable=W0611\n"
        f"{INDENT}BaseModel,\n"
        f"{INDENT}Extra,\n"
        f"{INDENT}Field,\n"
        f"{INDENT}Json,\n"
        ")"
    )

    CONF = (
        f"class {BASE_CLASS_NAME}(BaseModel):\n"
        f"{INDENT}class Config:\n"
        # https://pydantic-docs.helpmanual.io/usage/model_config/#smart-union
        # https://stackoverflow.com/a/69705356/4478420
        f"{INDENT}{INDENT}smart_union=True\n"
        f"{INDENT}{INDENT}extra=Extra.forbid"
    )


class PydanticV2Plugin(PydanticBase):
    """Pydantic v2 plugin for Qenerate."""

    HEADER = (
        '"""\nGenerated by qenerate plugin=pydantic_v2. DO NOT MODIFY MANUALLY!\n"""\n'
    )

    IMPORTS = (
        "from collections.abc import Callable  # noqa: F401 # pylint: disable=W0611\n"
        "from datetime import datetime  # noqa: F401 # pylint: disable=W0611\n"
        "from enum import Enum  # noqa: F401 # pylint: disable=W0611\n"
        "from typing import (  # noqa: F401 # pylint: disable=W0611\n"
        f"{INDENT}Any,\n"
        f"{INDENT}Optional,\n"
        f"{INDENT}Union,\n"
        ")\n"
        "\n"
        "from pydantic import (  # noqa: F401 # pylint: disable=W0611\n"
        f"{INDENT}BaseModel,\n"
        f"{INDENT}ConfigDict,\n"
        f"{INDENT}Field,\n"
        f"{INDENT}Json,\n"
        ")"
    )

    CONF = (
        f"class {BASE_CLASS_NAME}(BaseModel):\n"
        f"{INDENT}model_config = ConfigDict(\n"
        f"{INDENT}{INDENT}extra='forbid'\n"
        f"{INDENT})"
    )



import json
import os
from pathlib import Path
from typing import cast

from graphql import GraphQLSchema, IntrospectionQuery, build_client_schema

from qenerate.core.feature_flag_parser import FeatureFlagError
from qenerate.core.plugin import GeneratedFile, Plugin
from qenerate.core.preprocessor import GQLDefinition, GQLDefinitionType, Preprocessor
from qenerate.plugins.pydantic.plugin import (
    PydanticV1Plugin,
    PydanticV2Plugin,
)

plugins: dict[str, Plugin] = {
    "pydantic_v1": PydanticV1Plugin(),
    "pydantic_v2": PydanticV2Plugin(),
}


class CodeCommand:
    def __init__(
        self, preprocessor: Preprocessor, plugins: dict[str, Plugin] = plugins
    ) -> None:
        self._preprocessor = preprocessor
        self._plugins = plugins

    @staticmethod
    def _find_query_files(directory: str) -> list[str]:
        result: list[str] = []
        for root, _, files in os.walk(directory):
            result.extend(
                os.path.join(root, name) for name in files if name.endswith(".gql")
            )
        return result

    def _preprocess(self, directory: str, schema: GraphQLSchema) -> list[GQLDefinition]:
        definitions: list[GQLDefinition] = []
        for file in self._find_query_files(directory):
            try:
                definitions.extend(self._preprocessor.process_file(Path(file)))
            except FeatureFlagError:  # noqa: PERF203
                continue
        self._preprocessor.validate(
            definitions=definitions,
            schema=schema,
        )
        return definitions

    def generate_code(self, introspection_file_path: str, directory: str) -> None:
        introspection = json.loads(
            Path(introspection_file_path).read_text(encoding="utf-8")
        )["data"]

        schema = build_client_schema(cast("IntrospectionQuery", introspection))

        definitions = self._preprocess(
            directory=directory,
            schema=schema,
        )
        operations_by_plugin: dict[str, list[GQLDefinition]] = {
            plugin: [] for plugin in self._plugins
        }
        fragments_by_plugin: dict[str, list[GQLDefinition]] = {
            plugin: [] for plugin in self._plugins
        }

        for definition in definitions:
            plugin_name = definition.feature_flags.plugin
            if plugin_name not in self._plugins:
                # TODO: proper logging
                continue

            if definition.kind in {GQLDefinitionType.QUERY, GQLDefinitionType.MUTATION}:
                operations_by_plugin[definition.feature_flags.plugin].append(definition)
            elif definition.kind == GQLDefinitionType.FRAGMENT:
                fragments_by_plugin[definition.feature_flags.plugin].append(definition)

        generated_files: list[GeneratedFile] = []
        for plugin_name in self._plugins:
            plugin = self._plugins[plugin_name]

            rendered_fragments = plugin.generate_fragments(
                definitions=fragments_by_plugin[plugin_name],
                schema=schema,
            )

            rendered_queries = plugin.generate_operations(
                definitions=operations_by_plugin[plugin_name],
                fragments=rendered_fragments,
                schema=schema,
            )

            generated_files.extend(rendered_fragments)
            generated_files.extend(rendered_queries)

        for file in generated_files:
            file.save()



import re
from collections.abc import Mapping

from graphql import GraphQLOutputType


def _keyword_sanitizer(s: str) -> str:
    keyword_remapping = {
        "global": "q_global",
        "from": "q_from",
        "type": "q_type",
        "id": "q_id",
        "to": "q_to",
        "format": "q_format",
    }
    if s in keyword_remapping:
        return keyword_remapping[s]
    return s


def graphql_class_name_to_python(graphql_type: GraphQLOutputType) -> str:
    return graphql_class_name_str_to_python(str(graphql_type))


def graphql_class_name_str_to_python(class_name: str) -> str:
    result = class_name[0]
    for i in range(1, len(class_name)):
        cur, prev = class_name[i], class_name[i - 1]
        if cur != "_" and prev != "_":
            result += cur
        elif cur != "_" and prev == "_":
            result += cur.upper()
    return result


def graphql_primitive_to_python(
    graphql_type: GraphQLOutputType, custom_mappings: Mapping[str, str]
) -> str:
    key = str(graphql_type)
    if key in custom_mappings:
        return custom_mappings[key]
    pre_defined_mappings = {
        "ID": "str",
        "String": "str",
        "Int": "int",
        "Float": "float",
        "Boolean": "bool",
        "DateTime": "datetime",
        "JSON": "Json",
    }
    return pre_defined_mappings.get(key, "str")


def graphql_field_name_to_python(name: str) -> str:
    # ElbFQDN -> ["elb", "f", "q", "d", "n"]
    # SLOParameter -> ["s", "l", "o", "parameter"]
    parts = re.split(r"(?=[A-Z])", name)
    if not parts[0]:
        # if first letter was capital, then split will add '' first
        parts = parts[1:]
    for i, el in enumerate(parts):
        parts[i] = el.lower()

    # ["elb", "f", "q", "d", "n"] -> elb_fqdn
    # ["s", "l", "o", "parameter"] -> slo_parameter
    result = parts[0]
    for i in range(1, len(parts)):
        cur, prev = parts[i], parts[i - 1]
        if len(cur) == 1 and len(prev) == 1:
            result += cur
        else:
            result += f"_{cur}"
    result = result.lstrip("_")
    return _keyword_sanitizer(result)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from qenerate.core.preprocessor import GQLDefinitionType

BASE_CLASS_NAME = "ConfiguredBaseModel"
INDENT = "    "


@dataclass
class ParsedNode:
    parent: ParsedNode | None
    fields: list[ParsedNode]
    parsed_type: ParsedFieldType

    def class_code_string(self) -> str:  # noqa: PLR6301
        return ""

    def _needs_class_rendering(self) -> bool:
        if self.parsed_type.is_primitive:
            return False

        return not self._is_non_partial_fragment()

    def _is_non_partial_fragment(self) -> bool:
        return len(self.fields) == 1 and isinstance(
            self.fields[0], ParsedFragmentSpreadNode
        )

    @staticmethod
    def _empty() -> str:
        return f"{INDENT}..."


@dataclass
class ParsedInlineFragmentNode(ParsedNode):
    def class_code_string(self) -> str:
        # Assure not Optional[]
        if not (self.parent and self.parsed_type):  # type: ignore[truthy-bool]
            return ""

        if self.parsed_type.is_primitive:
            return ""

        base_class = self.parent.parsed_type.unwrapped_python_type
        if isinstance(self.parent, ParsedFragmentDefinitionNode):
            base_class = self.parent.fragment_name

        lines = ["\n\n"]
        lines.append(f"class {self.parsed_type.unwrapped_python_type}({base_class}):")
        fields_added = False
        for field in self.fields:
            if isinstance(field, ParsedClassNode):
                lines.append(
                    f"{INDENT}{field.py_key}: {field.field_type()} = "
                    f'Field(..., alias="{field.gql_key}")'
                )
                fields_added = True

        if not fields_added:
            lines.append(self._empty())

        return "\n".join(lines)


@dataclass
class ParsedClassNode(ParsedNode):
    gql_key: str
    py_key: str

    def class_code_string(self) -> str:
        if not self._needs_class_rendering():
            return ""

        if self.parsed_type.enum_map:
            return self._enum_code()
        return self._class_code()

    def _class_code(self) -> str:
        base_classes = ", ".join(self._base_classes())
        lines = ["\n\n"]
        lines.append(f"class {self.parsed_type.unwrapped_python_type}({base_classes}):")
        fields_added = False
        for field in self.fields:
            field_arg = "..., "
            if field.parsed_type.enum_map:
                field_arg = ""
            if isinstance(field, ParsedClassNode):
                lines.append(
                    f"{INDENT}{field.py_key}: {field.field_type()} = "
                    f'Field({field_arg}alias="{field.gql_key}")'
                )
                fields_added = True

        if not fields_added:
            lines.append(self._empty())

        return "\n".join(lines)

    def _enum_code(self) -> str:
        lines = ["\n\n"]
        lines.append(f"class {self.parsed_type.unwrapped_python_type}(Enum):")
        for k, v in self.parsed_type.enum_map.items():
            val = f'"{v}"' if isinstance(v, str) else v
            lines.append(f"{INDENT}{k} = {val}")
        return "\n".join(lines)

    def _base_classes(self) -> list[str]:
        base_classes: list[str] = []
        for field in self.fields:
            if not isinstance(field, ParsedFragmentSpreadNode):
                continue
            base_classes.append(field.parsed_type.unwrapped_python_type)
        if not base_classes:
            base_classes.append(BASE_CLASS_NAME)
        return base_classes

    def field_type(self) -> str:
        # This is a full (non-partial) fragment spread
        if len(self.fields) == 1 and isinstance(
            self.fields[0], ParsedFragmentSpreadNode
        ):
            return self.parsed_type.wrapped_python_type.replace(
                self.parsed_type.unwrapped_python_type,
                self.fields[0].parsed_type.unwrapped_python_type,
            )

        unions: list[str] = []
        # TODO: sorting does not need to happen on each call
        """
        Pydantic does best-effort matching on Unions.
        Declare most significant type first.
        This, smart_union and disallowing extra fields gives high confidence
        in matching.
        https://pydantic-docs.helpmanual.io/usage/types/#unions
        """
        self.fields.sort(key=lambda a: len(a.fields), reverse=True)
        unions.extend(
            field.parsed_type.unwrapped_python_type
            for field in self.fields
            if isinstance(field, ParsedInlineFragmentNode)
        )
        if len(unions) > 0:
            unions.append(self.parsed_type.unwrapped_python_type)
            return self.parsed_type.wrapped_python_type.replace(
                self.parsed_type.unwrapped_python_type, f"Union[{', '.join(unions)}]"
            )
        return self.parsed_type.wrapped_python_type


@dataclass
class ParsedOperationNode(ParsedNode):
    operation_type: GQLDefinitionType

    def class_code_string(self) -> str:
        lines = ["\n\n"]
        class_suffix = "QueryData"
        if self.operation_type == GQLDefinitionType.MUTATION:
            class_suffix = "MutationResponse"
        lines.append(
            f"class {self.parsed_type.unwrapped_python_type}{class_suffix}"
            f"({BASE_CLASS_NAME}):"
        )
        fields_added = False
        for field in self.fields:
            if isinstance(field, ParsedClassNode):
                lines.append(
                    f"{INDENT}{field.py_key}: {field.field_type()} = "
                    f'Field(..., alias="{field.gql_key}")'
                )
                fields_added = True

        if not fields_added:
            lines.append(self._empty())

        return "\n".join(lines)


@dataclass
class ParsedFragmentDefinitionNode(ParsedNode):
    class_name: str
    fragment_name: str

    def class_code_string(self) -> str:
        lines = ["\n\n"]
        lines.append(f"class {self.class_name}({BASE_CLASS_NAME}):")
        fields_added = False
        for field in self.fields:
            if isinstance(field, ParsedClassNode):
                lines.append(
                    f"{INDENT}{field.py_key}: {field.field_type()} = "
                    f'Field(..., alias="{field.gql_key}")'
                )
                fields_added = True

        if not fields_added:
            lines.append(self._empty())

        return "\n".join(lines)


@dataclass
class ParsedFragmentSpreadNode(ParsedNode):
    def class_code_string(self) -> str:  # noqa: PLR6301
        return ""


@dataclass
class ParsedFieldType:
    unwrapped_python_type: str
    wrapped_python_type: str
    is_primitive: bool
    enum_map: dict[str, Any]



import json

import requests
from graphql import get_introspection_query


class IntrospectionCommand:
    @staticmethod
    def introspection_query(url: str) -> None:
        query = get_introspection_query()
        request = requests.post(url, json={"query": query}, timeout=30)
        if request.status_code == requests.codes.ok:
            print(json.dumps(request.json(), indent=4))  # noqa: T201
            return
        raise Exception(f"Could not query {url}")  # noqa: TRY002



from dataclasses import dataclass
from enum import Enum

from graphql import (
    GraphQLEnumType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLOutputType,
    GraphQLScalarType,
)


class WrapperType(Enum):
    LIST = 1
    OPTIONAL = 2


@dataclass
class UnwrapperResult:
    wrapper_stack: list[WrapperType]
    inner_gql_type: GraphQLOutputType
    is_primitive: bool
    enum_map: dict[str, str]


class Unwrapper:
    """Unwrapper is responsible for unwrapping those lists and non-optionals.

    GraphQLOutputType can be nested in lists and non-optionals.
    """

    @staticmethod
    def unwrap(gql_type: GraphQLOutputType) -> UnwrapperResult:
        wrappers: list[WrapperType] = []
        if isinstance(gql_type, GraphQLNonNull):
            gql_type = gql_type.of_type
        else:
            wrappers.append(WrapperType.OPTIONAL)

        if isinstance(gql_type, GraphQLList):
            res = Unwrapper.unwrap(gql_type.of_type)
            wrappers.append(WrapperType.LIST)
            wrappers.extend(res.wrapper_stack)
            return UnwrapperResult(
                wrapper_stack=wrappers,
                inner_gql_type=res.inner_gql_type,
                is_primitive=res.is_primitive,
                enum_map=res.enum_map,
            )

        enum_map = {}
        if isinstance(gql_type, GraphQLEnumType):
            for k, v in gql_type.values.items():
                enum_map[k] = v.value
        return UnwrapperResult(
            wrapper_stack=wrappers,
            inner_gql_type=gql_type,
            is_primitive=isinstance(gql_type, GraphQLScalarType),
            enum_map=enum_map,
        )



import argparse
from importlib.metadata import version

from qenerate.core.code_command import CodeCommand
from qenerate.core.introspection_command import IntrospectionCommand
from qenerate.core.preprocessor import Preprocessor


def run() -> None:
    parser = argparse.ArgumentParser(prog="qenerate")
    parser.add_argument(
        "-v", "--version", action="version", version=version("qenerate")
    )

    subparsers = parser.add_subparsers(help="Supported commands", dest="subcommand")

    parser_introspection = subparsers.add_parser(
        "introspection",
        help="Run introspection query on given GQL URL.",
    )
    parser_introspection.add_argument("url", type=str, help="URL to run query against")

    parser_generator = subparsers.add_parser(
        "code",
        help="Generate Code",
    )
    parser_generator.add_argument(
        "-i",
        dest="introspection",
        type=str,
        help="Specify introspection query json",
    )
    parser_generator.add_argument(
        "dir", type=str, help="Specify introspection query json"
    )

    args = parser.parse_args()

    if args.subcommand == "introspection":
        IntrospectionCommand.introspection_query(args.url)
    elif args.subcommand == "code":
        code_command = CodeCommand(preprocessor=Preprocessor())
        code_command.generate_code(
            introspection_file_path=args.introspection,
            directory=args.dir,
        )


if __name__ == "__main__":
    run()
